//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalCommon.h	--- definitions common to all Cal constituents
//
//	DESCRIPTION:
//		This file contains declarations of basic classes, constants and
//		methods common to several Cal constituents.
//
//	CHANGE HISTORY:
//		UL, 1999-03-29: 1.0.0: - first version
//		UL, 1999-09-02: 1.1.0: - added BinnedAxis class family
//		UL, 1999-10-27: 1.2.0: - constructurs of NonUniformilyBinnedAxis
//								 take upper bound of last bin explicitely
//		UL, 2000-05-30: 1.3.0: - added new NonUniformilyBinnedAxis constructor
//
//	AUTHOR:
//		U. Lammers, Mon Mar 29 15:09:59 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/cal/src/CalCommon.h,v 1.30 2003/10/03 14:15:41 sasbuild Exp $
//
#ifndef CAL_COMMON_H
#define CAL_COMMON_H

#include <map>
#include <functional>
#include <algorithm>
#include <vector>

#include "common.h"
#include "Xmm.h"
#include "Ccf.h"
#include "CalDataVector.h"
#include "PhysicalConstants.h"
#include "Dal.h"
#include "slatec++.h"

//
//	Cal accuracy
//
typedef enum { LOW, MEDIUM, EXTENDED, HIGH } AccuracyLevel;
const std::string &toString(AccuracyLevel accuracyLevel);
extern const std::vector<AccuracyLevel> accuracyChoices;
namespace std {
std::ostream &operator<<(std::ostream &os, AccuracyLevel accuracyLevel);
};

//
//	general definitions
//
typedef pair<short, short>	PixelLocation;

typedef pair<double, double> FocalPlanePosition;

typedef pair<double, double> DPair;

class RealFocalPlanePosition {
public:
	RealFocalPlanePosition(double z = 0., double y = 0.) : mmZ(z), mmY(y) {}
	double	mmZ, mmY;
};

struct PolarFocalPlanePosition {
public:
	PolarFocalPlanePosition(double thetaIn = 0., double phiIn = 0.)
		: theta(thetaIn), phi(phiIn) {}
	double	theta, phi;
};

struct AsymmetricError {
	double	errorPlus, errorMinus;
};

struct ValueError {
	double	value;
	double	error;
};

struct ValueAsymmetricError {
	double	value;
	AsymmetricError	error;
};

class CCDwindow {
public:
	CCDwindow(short xmin = 1, short ymin = 1, short xmax = 1, short ymax = 1) :
			_xmin(xmin), _ymin(ymin), _xmax(xmax), _ymax(ymax) {}

	~CCDwindow() {}

	bool isPixelInside(const PixelLocation &pixel) const;

private:
	short _xmin, _ymin, _xmax, _ymax;
}; // CCDwindow


//
//	---( SpatialWindow )----------------------------------------------------
//
template<class Type>
class SpatialWindow {
public:
	SpatialWindow(Type halfX, Type halfY) : _halfX(halfX), _halfY(halfY) {}
	~SpatialWindow() {}

	Type halfX() const { return _halfX; }
	Type halfY() const { return _halfY; }

	void halfX(Type halfX) { _halfX = halfX; }
	void halfY(Type halfY) { _halfY = halfY; }

private:
	Type	_halfX, _halfY;
};


template<class Type, class ScalarType>
SpatialWindow<Type> operator*(const SpatialWindow<Type> &window, ScalarType s)
{
	return SpatialWindow<Type>(window.halfX() * s,
							   window.halfY() * s);
} // operator*


//
//	---( BinSize )----------------------------------------------------------
//
template<class Type>
class BinSize {
public:
	BinSize(Type binX = 0, Type binY = 0) : _binX(binX), _binY(binY) {}
	~BinSize() {}

	Type binX() const { return _binX; }
	Type binY() const { return _binY; }

	void binX(Type binX) { _binX = binX; }
	void binY(Type binY) { _binY = binY; }

private:
	Type	_binX, _binY;
};


template<class Type, class ScalarType>
BinSize<Type> operator*(const BinSize<Type> &binSize, ScalarType s)
{
	return BinSize<Type>(binSize.binX() * s, binSize.binY() * s);
} // operator*


//
//	---( Spectrum )----------------------------------------------------------
//
//
//	A Spectrum is a finite vector the index numbers of which signify 'channel'
//	numbers (starting at 0). The contents of a channel is the value of a
//	quantity at that channel number - usually a Spectrum represents a
//	probability distribution.

class Spectrum : public std::vector<double> {
public:
	Spectrum() : _momentSet(false) {}

	Spectrum(unsigned size) : std::vector<double>(size), _momentSet(false) {}

	Spectrum(unsigned size, double initialValue) : std::vector<double>(size),
	_momentSet(false) { operator=(initialValue); }

	Spectrum(const std::vector<double> &v) :
		std::vector<double>(v), _momentSet(false) {}

	Spectrum(std::vector<double>::const_iterator from,
			 std::vector<double>::const_iterator to)
		: std::vector<double>(from, to), _momentSet(false) {}
		//
		//	PURPOSE:
		//		Construct a Spectrum
		//			+ of size 0
		//			+ of given size uninitialized
		//			+ of given size initialized
		//			+ from contents of other vector
		//			+ range of iterators


	virtual Spectrum &operator=(double rhs);
		//
		//	PURPOSE:
		//		Set all values of spectrum to <rhs>
		//

	virtual Spectrum &operator=(const Spectrum &rhs);
		//
		//	PURPOSE:
		//		Assignment operator
		//

	virtual Spectrum &operator+=(const Spectrum &rhs);
		//
		//	PURPOSE:
		//		Addition of spectrum <rhs>


	virtual Spectrum &normalize(double norm = 1.);
		//
		//	PURPOSE:
		//		Normalize spectrum to given value such that
		//
		//		   size-1
		//			SUM	  v[i] = <norm>
		//			i=0


	virtual double norm() const;
		//
		//	PURPOSE:
		//		Inquire normalization:
		//		   size-1
		//			SUM	  v[i]
		//			i=0
		//


	virtual void rotate(int channels);
		//
		//	PURPOSE:
		//		Shift all channels by an amount <channels> in a wrap-around
		//		fashion, i.e., channels that are pushed out at the end
		//		are re-inserted at the beginning. Useful for convolutions.
		//


	virtual unsigned peakChannel() const;
		//
		//	PURPOSE:
		//		Returns channels number with largest value.
		//


	virtual void firstMoment(double moment1) { _momentSet = true;
											   _moment1 = moment1; }
	virtual double firstMoment() const;
		//
		//	PURPOSE:
		//		Return first moment of distribution, i.e.,
		//					 size-1
		//			<mom_1> = SUM   i*v[i]
		//					  i=0


	virtual ~Spectrum() {}

private:
	bool	_momentSet;
	double	_moment1;
};


template<class T>
class Interval {
public:
	Interval(T lo = 0, T hi = 0);	// default constructor
	T lo() const { return _lo; }
	T hi() const { return _hi; }
	double mid() const { return .5*(_lo + _hi); }
	T width() const { return _hi - _lo; }
	bool contains(T x) const { return x>=_lo && x<=_hi; }

private:
	T	_lo, _hi;
};

template<class T>
bool operator<(const Interval<T> &t1, const Interval<T> &t2)
{
	return t1.lo() < t2.lo() || (t1.lo()==t2.lo() && t1.hi() < t2.hi());
} // operator<


template<class T>
class IntervalVector : public std::vector<Interval<T> > {
public:
	//	in which interval is a value?
	//	requires:	vector needs to be sorted
	//	promise:	returns number of first interval which contains value
	//				or -1 if not found at all
	long whichInterval(const T &value) const;
};


template<class T>
Interval<T>::Interval(T lo, T hi)
{
	if (lo <= hi) {
		_lo = lo;
		_hi = hi;
	} else {
		_lo = hi;
		_hi = lo;
	}
} // Interval<T>::Interval()


template<class T>
long IntervalVector<T>::whichInterval(const T &value) const
{
	unsigned long size = this->size();
	if (size == 0 || value < (*this)[0].lo() || value > (*this)[size-1].hi())
		//	vector is empty or value is out of bounds
		return -1;

	long lo = 0,
		 hi = size - 1,
		 mid = (lo+hi)/2;
	while (lo < hi) {
        if (value < (*this)[mid].lo())
            hi = mid - 1;
        else if (value > (*this)[mid].hi())
            lo = mid + 1;
		else
			break;
        mid = (lo+hi)/2;
    }

	return	mid >= 0 &&
			value >= (*this)[mid].lo() &&
			value <= (*this)[mid].hi() ? mid : -1L;
} // IntervalVector<T>::whichInterval()


//
// --------------------( SupplementableVector )-------------------------------
//
template<class T>
class SupplementableVector : public std::vector<T> {
public:
	SupplementableVector() {}
	SupplementableVector(const std::vector<T> &v)
		: std::vector<T>(v) {}

	void supplement(const std::vector<T> supp)
	{
		typename std::vector<T>::const_iterator beg = supp.begin(),
					   vend = supp.end();
		for (typename std::vector<T>::const_iterator i = beg;
			 i != vend; ++i) {
			typename SupplementableVector<T >::iterator p =
												lower_bound(begin(),end(), *i);
			if (p==end() || *p!=*i)
				insert(p, *i);
		}
	}
};

typedef SupplementableVector<double>	SupplementableDoubleVector;


//
// --------------------( BinnedAxis )-----------------------------------------
//
typedef Interval<double> Bin;

//
//	BinnedAxis: abstract class modelling a binned real-valued
//				data axis; the axis extents from a lower bound to
//				an upper bound and consists of a fixed number of
//				- not necessarily equally wide - bins; the value of
//			    a bin represents the number of data points in it;
//				There is no space between adjacent bins and each bin is a
//				semi-open interval: the lower bound of a bin #i
//				belongs to it and equals the upper bound of bin #i-1 (which
//				does not belong to #i-1) => the upper bound of the last
//				bin of the axis does not belong to the axis
//
class BinnedAxis {
public:
	//	exception classes
	class NoSuchBin{};
	class IllegalInput{};
		//
		//	thrown by below methods if input arguments do not correspond
		//	to an actual bin
		//

	BinnedAxis();
	BinnedAxis(double lower, double upper);
		//
		//	PURPOSE:
		//		Construct a BinnedAxis object extending from a lower
		//		left edge <lower> to an upper right edge <upper>
		//

	double lower() const { return _lower; }
	double upper() const { return _upper; }
		//
		//	PURPOSE:
		//		Return the lower and upper edges of the axis object
		//

	virtual unsigned bins() const = 0;
		//
		//	PURPOSE:
		//		Returns the number of bins of the axis
		//

	virtual Bin bin(unsigned index = 0) const = 0;
		//
		//	PURPOSE:
		//		Returns the bin corresponding to a given index
		//			0<=index<MAX
		//		with MAX being the number of bins returned by above method;
		//		a NoSuchBin() exception will be thrown if the index is
		//		of of bounds
		//

	virtual Bin bin(double value) const = 0;
		//
		//	PURPOSE:
		//		Returns the bin corresponding to a given value along
		//		the axis; an NoSuchBin() exception will be thrown if
		//		the following condition is not met:
		//			MIN <= value < MAX
		//		with MIN/MAX being the return values of the above
		//		methods lower()/upper()
		//


	virtual unsigned index(double value) const = 0;
		//
		//	PURPOSE:
		//		Return the index of the bin whose lower bound is less than
		//		and upper bound greater than specified value
		//


	virtual unsigned &operator[](unsigned index) = 0;
		//
		//	PURPOSE:
		//		Access the contents of a bin by its index
		//			0<=index<MAX
		//		with MAX being the number of bins returned by above method
		//		bins(); a NoSuchBin() exception will be thrown if the index is
		//		of of bounds
		//

	virtual unsigned &operator[](double value) = 0;
		//
		//	PURPOSE:
		//		Access the contents of a bin through a value along the
		//		axis; an NoSuchBin() exception will be thrown if the
		//		following condition is not met:
		//			MIN <= value < MAX
		//		with MIN/MAX being the return values of the above
		//		methods lower()/upper()
		//

	virtual ~BinnedAxis();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor
		//
protected:
	double	_lower, _upper;
}; // BinnedAxis


//
//	UniformilyBinnedAxis: implements a BinnedAxis with uniform bins
//
class UniformilyBinnedAxis : public virtual BinnedAxis {
public:
	UniformilyBinnedAxis(double lower, double upper, unsigned bins);
		//
		//	PURPOSE:
		//		Construct a UniformilyBinnedAxis object
		//
		//	ARGUMENTS:
		//		lower/upper: left edge of the lowest/right edge of highest bin
		//		bins	   : number of bins
		//

	UniformilyBinnedAxis(double lower, unsigned bins, double binWidth);
		//
		//	PURPOSE:
		//		Construct a UniformilyBinnedAxis object
		//
		//	ARGUMENTS:
		//		lower      : left edge of the lowest
		//		bins	   : number of bins
		//		binWidth   : the bin width; the right edge of the highest
		//					 bin wil be lower+bins*binWidth
		//

	//
	//	see descriptions of corresponding BinnedAxis methods
	//
	unsigned bins() const;
	Bin bin(unsigned index = 0) const;
	Bin bin(double value) const;
	unsigned &operator[](unsigned index);
	unsigned &operator[](double value);
	unsigned index(double value) const;

	~UniformilyBinnedAxis();

private:
	double		_binWidth;
	unsigned	_bins;
	unsigned	*_v;
}; // UniformilyBinnedAxis


//
//	UniformilyBinnedAxis: implements a BinnedAxis with non-uniform bins
//
class NonUniformilyBinnedAxis : public virtual BinnedAxis {
public:
	NonUniformilyBinnedAxis();
	NonUniformilyBinnedAxis(const CalReal64Vector &lowerBounds,
							bool withUpperBound = false,
							real64 upperBound=0.);
	NonUniformilyBinnedAxis(const CalReal32Vector &lowerBounds,
							bool withUpperBound = false,
							real32 upperBound=0.f);
		//
		//	PURPOSE:
		//		Construct a NonUniformilyBinnedAxis object
		//
		//	ARGUMENTS:
		//		lowerBounds: vector with lower bounds of the bins;
		//					 the size has to be N+1 with N being
		//					 the desired number of bins and
		//					 lowerBounds[N] being the right edge of
		//					 the highest bin (which does not belong
		//					 to that bin)
		//		withUpperBound:
		//					 Is the upper bound of the last bin explicitely
		//					 given in last argument?
		//		upperBound:	 Upper bound of last bin if <withUpperBound>=true
		//


	//
	//	see descriptions of corresponding BinnedAxis methods
	//
	unsigned bins() const;
	Bin bin(unsigned index = 0) const;
	Bin bin(double value) const;
	unsigned &operator[](unsigned index);
	unsigned &operator[](double value);

	~NonUniformilyBinnedAxis();

private:
	template<class rvtype, class rtype> void init(const rvtype &, bool, rtype);

	std::vector<double>	_lowerBounds;
	unsigned		*_v;
	unsigned	_bins;
	unsigned	index(double value) const;
}; // NonUniformilyBinnedAxis

//
// -----( Logarithmic grid )---------------------------------------------------
//
template<class NumericType>
class LogarithmicGrid : public std::vector<NumericType> {
public:
	LogarithmicGrid(unsigned gridSize = 2, NumericType maxValue = 1.);
	~LogarithmicGrid() {}

private:
	NumericType value(unsigned) const;
	NumericType _i0;
};

typedef LogarithmicGrid<double>	LogarithmicDoubleGrid;

//
// -----( InterpolatedCurve / InterpolatedSurface )----------------------------
//
typedef Slatec::CubicHermiteInterpolant<double>	InterpolatedCurve;
typedef Slatec::Bivariate2DInterpolantRectangular<float> InterpolatedSurface;

//
// -----( CalFunctors )--------------------------------------------------------
//
namespace CalFunctors {

template<class T>
class Inverse {
public:
	T operator()(T v) { return 1./v; }
};

template<class T>
class Angstrom2eV {
public:
	T operator()(T v) { return PhysicalConstants::eVtimesAo/v; }
};

}

//
//	miscellaneous utilites
//
//	multiply vector elements by scalar factor
//
template<class VType, class MType>
std::vector<VType> &operator*=(std::vector<VType> &v, MType fac)
{
	unsigned N = v.size();
	for (unsigned n=0; n<N; v[n++] *= fac);
	return v;
}

namespace CalCommon {

template<class T>
bool sorted(const std::vector<T> &v) {
	return !v.empty() ? (std::find_if(v.begin(), v.end(),
						std::bind2nd(std::less<T>(), v[0])) == v.end()) : true;
}

}

#endif
