//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalState.h	--- several classes defining the CalState
//
//	DESCRIPTION:
//		Several classes are declared which define a CalState object.
//
//	CHANGE HISTORY:
//		UL, 1999-03-24: 1.0.0: - skeleton
//		UL, 1999-10-11: 1.1.0: - boundary checks on exposureStartTime
//								 changed; is now:
//									time >= Xmm::missionReferenceTime
//		UL, 2002-07-17: 1.2.0: - new variable 'exposureDuration'
//		UL, 2002-09-12: 1.3.0: - minimum exposureDuration reduced to 300s
//								 (was 1000s)
//
//	AUTHOR:
//		U. Lammers, Wed Mar 24 14:29:44 MET 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/cal/src/CalState.h,v 1.18 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef CALSTATE_H
#define CALSTATE_H

#include "CalCommon.h"
#include "StateVariable.h"	//	general Cal include file
#include "CalChild.h"		//	general Cal include file
#include "Xmm.h"			//	things about Xmm
#include "STime.h"			//	we need time


extern const std::vector<bool> boolChoices;

class CalServer;

//
//	CalState: define the state of the CalServer object by a set of
//			  variables
class CalState : public CalChild<CalServer> {
public:
	CalState(const CalServer *server = 0) : CalChild<CalServer>(server),
				 //
				 //	the following state variables are inactive in the
				 //	base clase - the instrument states specify active
				 //	versions
				 //
				 _ccd("ccd"),							//	CCD id
				 _node("node"),							//	node id
				 _mode("mode"),							//	mode id
				 _filter("filter"),						//	filter id
				 _ocb("ocb"),							//	OCB
				 _temperature("temperature"),			//	chip temperature
				 _temperatureCamera("temperatureCamera"),//	camera temperature

				 //
				 //	the following list of state variables are active
				 //	all the time
				 //
				 _exposureStartTime(Xmm::missionStartTime,
									   "exposureStartTime",
									   "none",
									   "The exposure start time",
									   true,
									   false,
									   Xmm::missionReferenceTime,
									   Xmm::missionEndTime),
														//	start time of
														//	exposure

				 _exposureDuration(30000.,
									  "exposure duration",
									  "s",
									  "The exposure duration",
									  true,
									  true,
									  300.,
									  XmmMission::orbitSimulator().
										orbitalElements().period),
														//	exposure length

							
				 _temperatureTracking(false,
									  "temperatureTracking",
									  "none",
									  "Should the CAL retrieve temperatures "
									  "of CCD/camera from HK data?",
									  boolChoices),
														//	temperature
														//	tracking switch
				 _randomize(true,
						    "randomize",
						    "none",
							"Shall integer pixel and channel numbers "
							"be randomized before being converted into "
							"real numbers?",
							 boolChoices),
														//	randomize chip
														//	coordinates?
				 _accuracyLevel(LOW,
								"accuracyLevel",
								"none",
								"level of accuracy of calibration entities",
								accuracyChoices),
														//	level of
														//	Cal accuracy

				 _auxiliaryParameter(0.,
								     "auxiliaryParameter",// name
									 "none",			//	no unit
									 "Auxiliary parameter",
									 false,
									 false)
		{}
		//
		//	PURPOSE:
		//		Create a new CalState object
		//

	virtual RangeStateVariable<Ccd> *ccd() const
		{ return &_ccd; }
	virtual ChoiceStateVariable<Node> *node() const
		{ return &_node; }
	virtual ChoiceStateVariable<Mode> *mode() const
		{ return &_mode; }
	virtual ChoiceStateVariable<Filter> *filter() const
		{ return &_filter; }
	virtual RangeStateVariable<unsigned> *ocb() const
		{ return &_ocb; }
	virtual RangeStateVariable<double> *temperature() const
		{ return &_temperature; }
	virtual RangeStateVariable<double> *temperatureCamera() const
		{ return &_temperatureCamera; }
	virtual ChoiceStateVariable<bool> *temperatureTracking() const
		{ return &_temperatureTracking; }
	virtual RangeStateVariable<STime> *exposureStartTime() const
		{ return &_exposureStartTime; }
	virtual RangeStateVariable<double> *exposureDuration() const
		{ return &_exposureDuration; }
	virtual ChoiceStateVariable<bool> *randomize() const
		{ return &_randomize; }
	virtual ChoiceStateVariable<AccuracyLevel> *accuracyLevel() const
		{ return &_accuracyLevel; }
	virtual RangeStateVariable<double> *auxiliaryParameter() const
		{ return &_auxiliaryParameter; }

	virtual ~CalState() {}
		//
		//	PURPOSE:
		//		destroy object constructed with above constructor

private:
	mutable RangeStateVariable<Ccd>				_ccd;
	mutable ChoiceStateVariable<Node>			_node;
	mutable ChoiceStateVariable<Mode>			_mode;
	mutable ChoiceStateVariable<Filter>			_filter;
	mutable RangeStateVariable<unsigned>		_ocb;
	mutable RangeStateVariable<double>			_temperature;
	mutable RangeStateVariable<double>			_temperatureCamera;
	mutable RangeStateVariable<STime>			_exposureStartTime;
	mutable RangeStateVariable<double>			_exposureDuration;
	mutable ChoiceStateVariable<bool>			_temperatureTracking;
	mutable ChoiceStateVariable<bool>			_randomize;
	mutable ChoiceStateVariable<AccuracyLevel>	_accuracyLevel;
	mutable RangeStateVariable<double>			_auxiliaryParameter;
}; // CalState


//
//	print CalState to a stream
//
std::ostream &operator<<(std::ostream &os, const CalState &calState);


//
//	----( CalStateEmos )-----------------------------------------------------
//
class CalStateEmos : public CalState {
public:
	CalStateEmos(const CalServer *server) : CalState(server),
					 _ccd(
						EMOScam::CCD_CENTRAL,	//	make central CCD def. one
						"ccd",					//	name of variable
						"none",					//	no unit
						"CCD number",			//	What is it?
						true,					//	there is a lowest number
						true,					//	there is a higest number
						EMOScam::CCD_FIRST,		//	lowest number
						EMOScam::CCD_LAST),		//	highest number

					 _node(
						EMOScam::PRIMARY_NODE,	//	default node
						"node",					//	name of variable
						"none",					//	no unit
						"node number",			//	What is it?
						xmmPayLoad.emos1.nodes()),//allowed nodes

					 _mode(
						xmmPayLoad.emos1.defaultMode(),	//	default mode
						"mode",					//	name of variable
						"none",					//	mode has no unit
						"mode identifier",		//	What is it?
						xmmPayLoad.emos1.modes()),	//	allowd modes

					 _filter(
						xmmPayLoad.emos1.defaultFilter(),//	default filter
						"filter",				//	name of variable
						"none",					//	no unit
						"filter identifier",	//	What is it?
						xmmPayLoad.emos1.filters()),// allowed filters

					 _temperature(
							120.,				//	120 K default
							"temperature",		//	name of variable
							"K",				//	Kelvin
							"temperature of the CCD", // What is it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.),				//	cannot get hotter than
												//	that
					 _temperatureCamera(
							120.,				//	120 K default
							"temperatureCamera",//	name of variable
							"K",				//	Kelvin
							"temperature of the camera structure",// Whatis it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.)				//	cannot get hotter than
												//	that
		{}

	RangeStateVariable<Ccd> *ccd() const
		{ return &_ccd; }
	ChoiceStateVariable<Node> *node() const
		{ return &_node; }
	ChoiceStateVariable<Mode> *mode() const
		{ return &_mode; }
	ChoiceStateVariable<Filter> *filter() const
		{ return &_filter; }
	RangeStateVariable<double> *temperature() const
		{ return &_temperature; }
	RangeStateVariable<double> *temperatureCamera() const
		{ return &_temperatureCamera; }

	~CalStateEmos() {}

private:
	mutable RangeStateVariable<Ccd>			_ccd;
	mutable ChoiceStateVariable<Node>		_node;
	mutable ChoiceStateVariable<Mode>		_mode;
	mutable ChoiceStateVariable<Filter>		_filter;
	mutable RangeStateVariable<double>		_temperature;
	mutable RangeStateVariable<double>		_temperatureCamera;
}; // CalStateEmos


//
//	----( CalStateEpn )-----------------------------------------------------
//
class CalStateEpn : public CalState {
public:
	CalStateEpn(const CalServer *server) : CalState(server),
					 _ccd(
						EPNcam::CCD_FIRST,		//	make first CCD default one
						"ccd",					//	name of variable
						"none",					//	no unit
						"CCD number",			//	What is it?
						true,					//	there is a lowest number
						true,					//	there is a higest number
						EPNcam::CCD_FIRST,		//	lowest number
						EPNcam::CCD_LAST),		//	highest number

					 _mode(
						xmmPayLoad.epn.defaultMode(),	//	default mode
						"mode",					//	name of variable
						"none",					//	mode has no unit
						"mode identifier",		//	What is it?
						xmmPayLoad.epn.modes()),//	allowd modes

					 _filter(
						xmmPayLoad.epn.defaultFilter(),//	default filter
						"filter",				//	name of variable
						"none",					//	no unit
						"filter identifier",	//	What is it?
						xmmPayLoad.epn.filters()),// allowed filters

					 _temperature(
							120.,				//	120 K default
							"temperature",		//	name of variable
							"K",				//	Kelvin
							"temperature of the CCD", // What is it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.),				//	cannot get hotter than
												//	room temperature
					 _temperatureCamera(
							120.,				//	120 K default
							"temperatureCamera",//	name of variable
							"K",				//	Kelvin
							"temperature of the camera structure", //Whatis it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.)				//	cannot get hotter than
												//	room temperature
		{}

	RangeStateVariable<Ccd> *ccd() const
		{ return &_ccd; }
	ChoiceStateVariable<Mode> *mode() const
		{ return &_mode; }
	ChoiceStateVariable<Filter> *filter() const
		{ return &_filter; }
	RangeStateVariable<double> *temperature() const
		{ return &_temperature; }
	RangeStateVariable<double> *temperatureCamera() const
		{ return &_temperatureCamera; }

	~CalStateEpn() {}

private:
	mutable RangeStateVariable<Ccd>			_ccd;
	mutable ChoiceStateVariable<Mode>		_mode;
	mutable ChoiceStateVariable<Filter>		_filter;
	mutable RangeStateVariable<double>		_temperature;
	mutable RangeStateVariable<double>		_temperatureCamera;
}; // CalStateEpn


//
//	----( CalStateRgs )-----------------------------------------------------
//
class CalStateRgs : public CalState {
public:
	CalStateRgs(const CalServer *server) : CalState(server),
					_ccd(
						RGScam::CCD_FIRST,		//	make first CCD default one
						"ccd",					//	name of variable
						"none",					//	no unit
						"CCD number",			//	What is it?
						true,					//	there is a lowest number
						true,					//	there is a higest number
						RGScam::CCD_FIRST,		//	lowest number
						RGScam::CCD_LAST),		//	highest number

					 _node(
						RGScam::PRIMARY_NODE,	//	default node
						"node",					//	name of variable
						"none",					//	no unit
						"node number",			//	What is it?
						xmmPayLoad.rgs1.nodes()),//	allowed nodes

					 _mode(
						xmmPayLoad.rgs1.defaultMode(),	//	default mode
						"mode",					//	name of variable
						"none",					//	mode has no unit
						"mode identifier",		//	What is it?
						xmmPayLoad.rgs1.modes()),			//	allowd modes

					 _ocb(
						RGScam::DEFAULT_OCB,	//	default OCBs
						"OCB",					//	name of the variable
						"none",					//	no unit
						"On Chip Binning factor", // What is it?
						true,					//	there is a lower limit
						true,					//	there is an upper limit
						1,						//	lower OCB limit
						5),						//	upper OCB limit

					 _temperature(
							120.,				//	120 K default
							"temperature",		//	name of variable
							"K",				//	Kelvin
							"temperature of the CCD", // What is it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.),				//	cannot get hotter than
												//	room temperature

					 _temperatureCamera(
							120.,				//	120 K default
							"temperatureCamera",//	name of variable
							"K",				//	Kelvin
							"temperature of the camera structure", //Whatis it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.)				//	cannot get hotter than
												//	room temperature
		{}

	RangeStateVariable<Ccd> *ccd() const
		{ return &_ccd; }
	ChoiceStateVariable<Node> *node() const
		{ return &_node; }
	ChoiceStateVariable<Mode> *mode() const
		{ return &_mode; }
	RangeStateVariable<unsigned> *ocb() const
		{ return &_ocb; }
	RangeStateVariable<double> *temperature() const
		{ return &_temperature; }
	RangeStateVariable<double> *temperatureCamera() const
		{ return &_temperatureCamera; }

	~CalStateRgs() {}

private:
	mutable RangeStateVariable<Ccd>			_ccd;
	mutable ChoiceStateVariable<Node>		_node;
	mutable ChoiceStateVariable<Mode>		_mode;
	mutable RangeStateVariable<unsigned>	_ocb;
	mutable RangeStateVariable<double>		_temperature;
	mutable RangeStateVariable<double>		_temperatureCamera;
}; // CalStateRgs


//
//	----( CalStateOm )-----------------------------------------------------
//
class CalStateOm : public CalState {
public:
	CalStateOm(const CalServer *server) : CalState(server),
					_ccd(
						OMcam::CCD_FIRST,		//	the is only one CCD
						"ccd",					//	name of variable
						"none",					//	no unit
						"CCD number",			//	What is it?
						true,					//	there is a lowest number
						true,					//	there is a higest number
						OMcam::CCD_FIRST,		//	lowest number
						OMcam::CCD_LAST),		//	highest number

					 _node(
						OMcam::PRIMARY_NODE,	//	default node
						"node",					//	name of variable
						"none",					//	no unit
						"node number",			//	What is it?
						xmmPayLoad.om.nodes()),	//	allowed nodes

					 _mode(
						xmmPayLoad.om.defaultMode(),//	default mode
						"mode",					//	name of variable
						"none",					//	mode has no unit
						"mode identifier",		//	What is it?
						xmmPayLoad.om.modes()),	//	allowd modes

					 _filter(
						xmmPayLoad.om.defaultFilter(),//	default filter
						"filter",				//	name of variable
						"none",					//	no unit
						"filter identifier",	//	What is it?
						xmmPayLoad.om.filters()),// allowed filters

					 _temperature(
							120.,				//	120 K default
							"temperature",		//	name of variable
							"K",				//	Kelvin
							"temperature of the CCD", // What is it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.),				//	cannot get hotter than
												//	room temperature

					 _temperatureCamera(
							120.,				//	120 K default
							"temperatureCamera",//	name of variable
							"K",				//	Kelvin
							"temperature of the camera structure",// Whatis it?
							true,				//	there is a lower limit
							true,				//	there is an upper limit
							0.,					//	no negative temperatures
							300.)				//	cannot get hotter than
												//	room temperature
		{}

	RangeStateVariable<Ccd> *ccd() const
		{ return &_ccd; }
	ChoiceStateVariable<Node> *node() const
		{ return &_node; }
	ChoiceStateVariable<Mode> *mode() const
		{ return &_mode; }
	ChoiceStateVariable<Filter> *filter() const
		{ return &_filter; }
	RangeStateVariable<double> *temperature() const
		{ return &_temperature; }
	RangeStateVariable<double> *temperatureCamera() const
		{ return &_temperatureCamera; }

	~CalStateOm() {}

private:
	mutable RangeStateVariable<Ccd>			_ccd;
	mutable ChoiceStateVariable<Node>		_node;
	mutable ChoiceStateVariable<Mode>		_mode;
	mutable ChoiceStateVariable<Filter>		_filter;
	mutable RangeStateVariable<double>		_temperature;
	mutable RangeStateVariable<double>		_temperatureCamera;
}; // CalStateOm

#endif
