//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CcdQuantumDataServer --- the CcdQuantumDataServer class
//
//	DESCRIPTION:
//		A class 'CcdQuantumDataServer' is defined which provides access to
//		quantum efficiency data.
//
//	CHANGE HISTORY:
//		UL, 2000-02-17: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Thu Feb 17 13:47:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbool, C. Erd et al.
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/cal/src/CcdQuantumDataServer.h,v 1.22 2003/10/03 14:15:41 sasbuild Exp $
//
#ifndef CCD_QUANTUM_DATA_SERVER_H
#define CCD_QUANTUM_DATA_SERVER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalCommon.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Evaluator.h"

//
//	---( CcdQuantumEfficiencyCurve )-------------------------------------------
//
class CcdQuantumEfficiencyCurve :
							public Slatec::CubicHermiteInterpolant<double>{
public:
	CcdQuantumEfficiencyCurve() {}
	CcdQuantumEfficiencyCurve(double value)
		: Slatec::CubicHermiteInterpolant<double>(value) {}
	CcdQuantumEfficiencyCurve(const std::vector<double> &eGrid,
							  const std::vector<double> &effvec)
		: Slatec::CubicHermiteInterpolant<double>(eGrid, effvec) {}


	CcdQuantumEfficiencyCurve &operator+=(const CcdQuantumEfficiencyCurve &);
		//
		//	PURPOSE:
		//		Add a QE curve to the current one
		//


	virtual double efficiency(double E) const {
		return operator()(E);
	}
		//
		//	PURPOSE:
		//		Evaluate CCD quantum efficiency curve at a single energy <E>
		//
		//	ARGUMENTS:
		//		E	: energy in eV
		//
		//	RETURNS:
		//		Value of CCD quantum efficiency curve - unitless number in
		//		range [0, 1]
		//


	virtual double efficiency(double Elo, double Ehi) const;
		//
		//	PURPOSE:
		//		Return weighted integral of CCD quantum efficiency curve
		//		between the two energies
		//
		//	ARGUMENTS:
		//		Elo, Ehi	: energies in eV between which quantum efficiency
		//					  is to be integrated; Elo < Elo
		//
		//	RETURNS:
		//
		//				 Ehi
		//		   1	/
		//		------	| dE qe(E)
		//		Ehi-Elo	/
		//				Elo
		//

	~CcdQuantumEfficiencyCurve() { }
};
typedef CcdQuantumEfficiencyCurve CcdLargePatternCurve;

/*
class CcdLargePatternCurve : public CcdQuantumEfficiencyCurve
{
   public:
        CcdLargePatternCurve() {}
	CcdLargePatternCurve(double value)
	        : CcdQuantumEfficiencyCurve(double value) {}
	CcdLargePatternCurve(const std::vector<double> &eGrid,
			      const std::vector<double> &effvec)
	        : CcdQuantumEfficiencyCurve(const std::vector<double> &eGrid,
		            const std::vector<double> &effvec) {}

        ~CcdLargePatternCurve() {}
};
*/

//
//	---( CcdQuantumDataServer )----------------------------------------
//
class PatternFractionDataServer;	//	forward declaration

class CcdQuantumDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<CcdQuantumDataServer> > {
public:
	CcdQuantumDataServer(Ccd ccd,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate CcdQuantumDataServer object applicable to CCD
		//		<ccd>
		//


	virtual void dump(ostream &) { }
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { return Ccf::QuantumEf; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual unsigned patterns() const { return 0; }
		//
		//	PURPOSE:
		//		Returns maximum number of patterns held by the server
		//


	virtual const CcdQuantumEfficiencyCurve &pattern(unsigned patternId)
		const = 0;
	virtual CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const = 0;
		//
		//	PURPOSE:
		//		Select a quantum efficiency curve for a particular
		//		pattern/grade
		//
		//	ARGUMENTS:
		//		gradeNr			: grade (number of pixels above threshold)
		//		patternId		: pattern ID ([0, 31] for EMOS)
		//


	virtual const CcdQuantumDataServer &at(unsigned short chipX,
										   unsigned short chipY) const = 0;
		//
		//	PURPOSE:
		//		Return quantum efficiency curve applicable at CCD pixel
		//		position (chipX, chipY) for pattern specified with preceding
		//		invocation of method pattern()
		//
		//	ARGUMENTS:
		//		chipX, chipY	: pixel position in CHIPCOORD frame at which
		//						  efficiency curve is requested
		//


	virtual const Proxy<PatternFractionDataServer> patternFractionData() const;
		//
		//	PURPOSE:
		//		Return an object that can retrieve explicit pattern fraction
		//		data from the CCF (EPIC only)
		//

//        virtual const double largeTo13_25() const=0;
                //
                //      PURPOSE: returns the ratio of large events to 
                //               pattern 13 to 25 events

	virtual const CcdLargePatternCurve sizeofLargePatterns() const=0;
                //
                //      PURPOSE: returns the average size of large patterns
                //               as a function of energy

	virtual ~CcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members
	Ccd				_ccd;
}; // CcdQuantumDataServer


//
//	instrument specific CcdQuantumDataServer
//	
//	EPIC (EMOS/EPN)
//
template<class T>
class QuantumDataRepository;	//	forward

class EpicCcdQuantumDataServer : public CcdQuantumDataServer {
public:
	EpicCcdQuantumDataServer(Ccd ccd,
							 const Table *table,
							 const Proxy<std::vector<double> > &eGrid,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate EpicCcdQuantumDataServer object from energy grid
		//		vector and corresponding Efficiency data
		//


	unsigned patterns() const { return _patterns; }
		//
		//	PURPOSE: see above
		//


	const CcdQuantumEfficiencyCurve &pattern(unsigned patternId) const;
		//
		//	PURPOSE: see above
		//
	

	CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const;
		//
		//	PURPOSE: see above
		//
	

	const CcdQuantumDataServer &at(unsigned short chipX,
								   unsigned short chipY) const;
		//
		//	PURPOSE: see above
		//


	const Proxy<PatternFractionDataServer> patternFractionData() const;
		//
		//	PURPOSE: see above
		//

	CcdLargePatternCurve getLargePatterns(const Table *table, 
                                   Proxy<vector<double> >eGrid);
		//
		//	PURPOSE: creates a CcdLargePatternCurve object
                //               from the large pattern keywords
		//
	const CcdLargePatternCurve sizeofLargePatterns() const; 
                     // {std::vector<double> e[] = {0.0,0.0}; return e;};
                //
                //      PURPOSE: see above
                //               
                                                                                
        // const double largeTo13_25() const;
                //
                //      PURPOSE: see above
                //
                                                                                
	virtual ~EpicCcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	unsigned						_patterns;		//	max number of patterns
	std::vector<std::vector<unsigned short> >
									_gradePatterns;	//	_gradePatterns[i] =
													//	list of all grade-i
													//	patterns

	//	native pattern QEs
	QuantumDataRepository<CcdQuantumEfficiencyCurve>	*_patternQECurves;
	CcdLargePatternCurve	_largePatternCurve;
	mutable Proxy<PatternFractionDataServer>			_pattFracServer;
        bool _largeSizePresent;
        Proxy<std::vector<double> >                                     _eGrid;
}; // EpicCcdQuantumDataServer

namespace CombinedEventGrades {
	//	possible pattern grades
	enum ID { INVALID	= -1,
			  SINGLE	= 0,	//	singles
			  DOUBLE	= 1,	//	doubles
			  TRIPLE	= 2,	//	triples,
			  QUADRUPLE	= 3,	//	quadruples
			  SD		= 4,	//	singles+doubles
			  SDTQ		= 5,	//	singles+doubles+triples+quadr.
			  // derived
			  LOWEST	= SINGLE,
			  HIGHEST	= SDTQ,
			  HIGHEST_PURE_GRADE = QUADRUPLE,
			  ALL_GRADES= HIGHEST
	};
	const char *const names[] = {
		"singles", "doubles", "triples", "quadruples", "s+d", "s+d+t+q" };
	bool contains(ID id, unsigned short grade);
};

typedef CombinedEventGrades::ID		EventGrades;

//
//	PatternFractionDataServer
//
class PatternFractionDataServer {
public:

	const CcdQuantumEfficiencyCurve &total() const;
		//
		//	PURPOSE:
		//		Return total quantum efficiency curve
		//


	const PatternFractionDataServer &at(unsigned short chipX,
										unsigned short chipY) const;
		//
		//	PURPOSE:
		//		Register position in CHIPCOORD coordinates at which pattern
		//		fraction data is requested
		//


	const PatternFractionDataServer &grade(EventGrades grades)
		const;
		//
		//	PURPOSE:
		//		Select pattern grade for which fraction data is to be
		//		requested with subsequent invocations of
		//			o inEnergySpace()
		//			o inChannelSpace()
		//


	const CcdQuantumEfficiencyCurve &inEnergySpace() const;
		//
		//	PURPOSE:
		//		Return pattern fraction curve in energy space for previously
		//		selected grade and chip position (see grade()/at() above)
		//


	const std::vector<double> &inChannelSpace() const;
	const std::vector<double> &inChannelSpace(std::vector<double> eGrid) const;
		//
		//	PURPOSE:
		//		Return pattern fraction data in channel space for previously
		//		selected grade and chip position (see grade()/at() above);
		//		second form re-evaluates the channel fraction data on new
		//		given energy grid <eGrid>
		//


	~PatternFractionDataServer();
		//
		//	Destroy object constructed with above constructor(s)	
		//

private:
	friend class EpicCcdQuantumDataServer;
	friend class CcdQuantumDataServer;
	explicit PatternFractionDataServer();

	//	pattern fraction data ...
	//
	//	... in energy space
	QuantumDataRepository<CcdQuantumEfficiencyCurve>	*_fracQECurves;
	class QEDataVector : public std::vector<double> {
	public:
		QEDataVector() : std::vector<double>() {}
		QEDataVector(const std::vector<double> &eGrid,
					 const std::vector<double> &v)
			: vector<double>(v), _eGrid(eGrid) {}

		const std::vector<double> &eGrid() const { return _eGrid; }
		void rebin(const std::vector<double> &eGrid) {
			Slatec::LinearInterpolant<double> li(_eGrid, *this);
			_eGrid.assign(eGrid.begin(), eGrid.end());
			unsigned N = eGrid.size();
			std::vector<double> newVals(N);
			for (unsigned n=0; n<N; ++n)
				newVals[n] = li(eGrid[n]);
			assign(newVals.begin(), newVals.end());
		}
	private:
		std::vector<double>	_eGrid;
	};

	//	... in channel space
	QuantumDataRepository<QEDataVector>			*_fracQEvectors;

	CcdQuantumEfficiencyCurve					_total;
	mutable EventGrades							_grade;

	//	mode
	Mode										_mode;

	//	methods
	std::pair<unsigned short, unsigned short> indexRange(const Table *tab,
											 unsigned short modeIndex) const;
	void checkGrade() const;
}; // PatternFractionDataServer


//
//	RGS
//
class RgsCcdQuantumDataServer : public CcdQuantumDataServer {
public:
	RgsCcdQuantumDataServer(Ccd ccd,
						 const Table *materialThicknessTable,
						 const Table *SiThicknessTable,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate RgsCcdQuantumDataServer object
		//


	const CcdQuantumEfficiencyCurve &pattern(unsigned patternId) const;
	CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const;
		//
		//	PURPOSE:
		//		Select a quantum efficiency curve for a particular
		//		pattern/grade
		//
		//	ARGUMENTS:
		//		gradeNr			: grade (number of pixels above threshold)
		//		patternId		: pattern ID ([0, 31] for Rgs)
		//


	const CcdQuantumDataServer &at(unsigned short chipX,
								   unsigned short chipY) const;
		//
		//	PURPOSE:
		//		Return quantum efficiency curve applicable at CCD pixel
		//		position (chipX, chipY) for pattern specified with preceding
		//		invocation of method pattern()
		//
		//	ARGUMENTS:
		//		chipX, chipY	: pixel position in CHIPCOORD frame at which
		//						  efficiency curve is requested
		//

//        const double largeTo13_25() const {return 0;};
	const CcdLargePatternCurve sizeofLargePatterns() const; 
                     // {std::vector<double> e[] = {0.0,0.0}; return e;};

	virtual ~RgsCcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	class ThicknessSiInterpolator {
	public:
		ThicknessSiInterpolator(double thickness)
			//			 um->cm
			: _thickness(1.e-4*thickness), _thicknessMap(0) { }
		ThicknessSiInterpolator(unsigned N, int16 *chipX, int16 *chipY,
								int16 *fringes, double d)
		{
			std::vector<float> x(N), y(N), v(N);
			for (unsigned n=0; n<N; ++n) {
				x[n] = FLT(chipX[n]);
				y[n] = FLT(chipY[n]);
				v[n] = 1.e-4*(d - FLT(fringes[n])/7.); // [2], Sect. 3.5.22.1.1
			}
			_thicknessMap = new Slatec::Bivariate2DInterpolantScattered<float>(
											x, y, v);
		}
		double operator()(int16 chipX, int16 chipY)
		{
			return _thicknessMap ? (*_thicknessMap)(FLT(chipX), FLT(chipY))
								 : _thickness;
		}
		~ThicknessSiInterpolator() { delete _thicknessMap; }
	private:
		double											_thickness;
		Slatec::Bivariate2DInterpolantScattered<float>  *_thicknessMap;
	};

	class RgsEXAFSDataServer {
	public:
		RgsEXAFSDataServer(Ccd ccd);

		bool hasMaterial(const std::string &name, bool isPassiveLayer) const;
		const RgsEXAFSDataServer &material(const std::string &mat,
										   bool isPassiveLayer) const;

		const std::vector<std::vector<double> > &E() const;
		const std::vector<std::vector<double> > &invMu() const;

		~RgsEXAFSDataServer() {}

	private:
		mutable int									_materialSelected;
		std::vector<std::pair<std::string, bool> >	_knownMaterials;
		std::vector<std::vector<std::vector<double> > >	_eGrids,
														_invMu;

		void checkMaterialSelected() const;
	};

	ThicknessSiInterpolator				*_thicknessSi;
	SupplementableDoubleVector			_eGrid;
	InterpolatedCurve					_transmission,
										_tau;

	mutable double						_currentSiThickness;

	//	methods
	void averageSiThickness() const;
}; // RgsCcdQuantumDataServer

#endif
