//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	GeometryDataServer --- classes to perform coordinate transformations
//
//	DESCRIPTION:
//		Several classes are declared which provide functionality to
//		convert coordinates from one frame into another one and vice
//		versa. The following frames are available:
//
//			PIXCOORD0: CCD pixel positions in telemetry format not corrected
//					   for under/overscan/windowing; start at 0; node-oriented;
//					   as available from ODF
//			PIXCOORD1: CCD pixel positions corrected for under/overscan and
//					   anti-window'ed; start at 1; node-oriented
//			CHIPCOORD: PIXCOORD1 for node 0 extending over the entire chip;
//					   no-longer node oriented
//			CAMCOORD1: physical positions in Cartesian camera coordinates;
//					   origin is physical center of camera
//			CAMCOORD2: physical positions in Cartesian camera coordinates;
//					   origin is location of telescope focus
//			SACCOORD:  S/C reference frame; Cartesian S/C coordinates;
//					   origin is in mirror support platform
//			TELCOORD:  telescope coordinates; angular coordinates of
//					   point in focal plane;
//			ROWCOORD:  Rowland coordinates; only applicable for RGS:
//					   two angles determine the position of an event
//					   on the RFC strips
//
//		The following conversions are provided:
//
//		PIXCOORD0->PIXCOORD1<->CHIPCOORD<->CAMCOORD1<->CAMCOORD2<->SACCOORD
//														  ^           ^
//                                                        |           |
//											              V           V
//                                                     TELCOORD    ROWCOORD
//
//	CHANGE HISTORY:
//		UL, 1999-07-05: 1.0.0: - first version from CalAtomsCommon
//		UL, 1999-09-06: 1.1.0: - added support for fractional CHIPCOORD pixels
//		UL, 2000-02-17: 1.2.0: - added mm2SacCoord() component to CFQ
//		UL, 2000-05-12: 1.3.0: - fixed minor problem in I/F
//		UL, 2001-02-21: 1.4.0: - added
//								  XmmCartesianPointList::rowlandDefocusLength()
//		UL, 2001-03-08: 1.5.0: - new methods:
//									XmmCartesianPointList::toChipCoord()
//									XmmCartesianPointList::onCcd()
//		UL, 2001-04-27: 1.6.0; - new methods:
//								   GeometryDataServer::lsfDefocusDistribution()
//								   XmmRowlandPointList::defocusLength()
//
//	AUTHOR:
//		U. Lammers (UL), Mon Jul  5 15:52:04 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/cal/src/GeometryDataServer.h,v 1.17 2003/10/03 14:15:41 sasbuild Exp $
//

#ifndef GEOMETRY_DATA_SERVER_H
#define GEOMETRY_DATA_SERVER_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"
#include "Celestial.h"
#include "Evaluator.h"
#include "Proxy.h"
#include "CalCommon.h"
#include "Xmm.h"
#include "Region.h"

#include <vector>

//
//	---( PixelList)--------------------------------------------------
//
template<class IntType>
class PixelList {
public:
	PixelList() {}
	PixelList(CalIntVector<IntType> &x,
			  CalIntVector<IntType> &y) : _x(x), _y(y) {}

	CalIntVector<IntType> &x() { return _x; }
	CalIntVector<IntType> &y() { return _y; }

	void shiftX(IntType dx = 0);
	void shiftY(IntType dy = 0);

	~PixelList() {}

protected:
	CalIntVector<IntType>	_x, _y;
}; // PixelList


//
//	---( CartesianPointList )-----------------------------------------------
//
class CartesianPointList {
public:
	CartesianPointList(CalReal64Vector &x,
					   CalReal64Vector &y,
					   CalReal64Vector &z) : _x(x), _y(y), _z(z) {}

	CalReal64Vector &x() { return _x; }
	CalReal64Vector &y() { return _y; }
	CalReal64Vector &z() { return _z; }

	virtual void translate(const CartesianVector &offset);
	virtual void rotate(const AttitudeMatrix &matrix);

	virtual ~CartesianPointList() {}

protected:
	CalReal64Vector	_x, _y, _z;
}; // CartesianPointList


//
//	---( AngularPointList )-----------------------------------------------
//
template<class T>
class AngularPointList {
public:
	AngularPointList(T &theta, T &phi) : _theta(theta), _phi(phi) {}

	T &theta() { return _theta; }
	T &phi() { return _phi; }

	virtual ~AngularPointList() {}

private:
	T	_theta, _phi;
}; // AngularPointList


//
//	---( XmmPixelList )-----------------------------------------------
//
struct CoordFrameQuantities {
	double			plateScaleX,			//	in mm/pixel
					plateScaleY,			//	in mm/pixel
					plateScaleLastY,		//	can be different for last pixel
											//	row (EPN)
					arcsecsPerMMx,			//	"/mm in X
					arcsecsPerMMy;			//	"/mm in Y
	unsigned		maxX,					//	maximum X in PIXOORD1
					maxY;					//	maximum Y in PIXCOOD1

	AttitudeMatrix	chipToCamCoord1Matrix,	//	CHIPCOORD -> CAMCOORD1
					cc2ToSacCoordMatrix,	//	CAMCOORD2 -> SACOORD
					sacCoordToRowCoordMatrix;
											//	SACCOORD -> ROWCOORD
	double			gry;

	CartesianVector	chipToCamCoord1Trans,	//	CHIPCOORD -> CAMCOORD1 transl.
					cc2ToSacCoordTrans,		//	CAMCOORD1 -> SACCOORD transl.
					cc1ToCc2,				//	CAMCOORD1 -> CAMCOORD2 transl.
					opticsToSacCoordTrans,	//	location of optical axis/
											//	focal plane intersection point
											//	in SACCOORD
					MM0,					//	center of mirror module ...
					G, M;					//	... G-point/M-point in SACOORD
	double			alpha;					//	alpha angle

	PixelLocation	centralPixel;			//	central pixel on chip

	double			refChipT, refCamT;		//	reference temperatures
											//	of chip + camera frame
	double			expansionChip,
					expansionCam;			//	expansionCoefficients of
											//	chip/camera in 1/K

	EulerAngles		sacCoordToCc2;			//	3-1-3 Euler angles for
											//	SACOORD->CAMCOORD2 transf.
}; //  CoordFrameQuantities()


class XmmCartesianPointList;
class GeometryDataServer;
class XmmRealPixelList;

class XmmPixelList : public PixelList<int16>,
					 public virtual CalChild<GeometryDataServer> {
public:
	enum CoordFrame { PIXCOORD0, PIXCOORD1, CHIPCOORD };
	XmmPixelList() : CalChild<GeometryDataServer>(0) {}
	XmmPixelList(CalInt16Vector &x,
				 CalInt16Vector &y,
				 CoordFrame frame = CHIPCOORD,
				 GeometryDataServer *parent = 0);

	//	real-valued pixels
	XmmRealPixelList toReal();

	//	PIXCOORD0 -> PIXCOORD1
	virtual XmmPixelList &toPixCoord1();
	//	PIXCOORD1 -> CHIPCOORD
	virtual XmmPixelList &toChipCoord(const CalInt8Vector *nodes = 0);
	//	CHIPCOORD -> CAMCOORD1
	virtual XmmCartesianPointList toCamCoord1(real64 *x = 0,
											  real64 *y = 0,
											  real64 *z = 0);
	virtual XmmCartesianPointList toCamCoord1(CalReal64Vector &x,
											  CalReal64Vector &y,
											  CalReal64Vector &z);

	//	CHIPCOORD -> PIXCOORD1
	virtual XmmPixelList &toPixCoord1(CalInt8Vector &nodes);


	virtual std::vector<bool> insideFOV();
		//
		//	PURPOSE:
		//		Determine for all pixels in the list whether they are
		//		within the FOV or not; return result as boolean vector
		//


	virtual ~XmmPixelList() {}

private:
	double xDoubleCoord(unsigned index) const;
	virtual double toDoubleX(unsigned index) const;
	double yDoubleCoord(unsigned index) const;
	virtual double toDoubleY(unsigned index) const;

	CoordFrameQuantities	_cfq;
	CartesianVector			_r0;
	unsigned				_ocbX, _ocbY;
	double					_randomize,
							_ocbrX, _ocbrY,
							_excessPixelsX,
							_excessPixelsY,
							_halfCentralOcbCeilX,
							_halfCentralOcbCeilY,
							_chipExpansion,
							_camExpansion;

protected:
	CoordFrame				_frame;
}; // XmmPixelList()


class XmmRealPixelList : public XmmPixelList {
public:
	XmmRealPixelList(CalReal64Vector &xr,
					 CalReal64Vector &yr,
					 CoordFrame frame = CHIPCOORD,
					 GeometryDataServer *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate a real-valued pixel list from two real
		//		vectors <xr>/<yr>; after successful construction
		//		the coordinates are broken into integer and fractional
		//		part the former of which is accessible through the
		//		x()/y() methods of the base class XmmPixelList. The
		//		fractional parts are accessed through xfrac()/yfrac() below;
		//		0<=xfrac[i]<1 / 0<=yfrac[i]<1
		//

	XmmRealPixelList(CalInt16Vector &x,
					 CalInt16Vector &y,
					 CoordFrame frame = CHIPCOORD,
					 GeometryDataServer *parent = 0);
		//
		//	PURPOSE:
		//		Some as above except that integer components
		//		coordinates are explicitely given
		//


	CalReal64Vector &xfrac() { return _xr; }
	CalReal64Vector &yfrac() { return _yr; }
		//
		//	PURPOSE:
		//		Access the fractional parts of the X/Y coordinates
		//

	real64 xr(unsigned index) const;
	real64 yr(unsigned index) const;
		//
		//	PURPOSE:
		//		Return the real-valued x/y coordinate of the pixel
		//		with index <index>

	void add(const CalReal64Vector &dx, const CalReal64Vector &dy);
		//
		//	PURPOSE:
		//		Add offset vectors <dx>/<dy> to all pixels;
		//		Please note: The elements of <dx>/<dy> are unlimited;
		//		Invokes normalize() below.
		//

	void normalize();
		//
		//	PURPOSE:
		//		Break the real-valued coordinates (which after construction
		//		are normalized to [0, 1) newly into integer and fractional
		//		part. Please note: This becomes necessary if the coordinates
		//		are directly modified via the vectors passed to the
		//		constructor.
		//

	~XmmRealPixelList() {}
		//
		//	PURPOSE:
		//		Destroy object instantiated with constructor above
		//

private:
	CalReal64Vector	_xr, _yr;

	void init();
	double toDoubleX(unsigned index) const;
	double toDoubleY(unsigned index) const;
}; // XmmRealPixelList


//
//	---( XmmCartesianPointList )-----------------------------------------------
//
class XmmRowlandPointList;
class XmmTelescopePointList;
class XmmCartesianPointList : public CartesianPointList,
							  public virtual CalChild<GeometryDataServer> {
public:
	enum CoordFrame { CAMCOORD1, CAMCOORD2, SACCOORD };
	XmmCartesianPointList(CalReal64Vector &x,
						  CalReal64Vector &y,
						  CalReal64Vector &z,
						  CoordFrame frame = CAMCOORD1,
						  GeometryDataServer *parent = 0);

	XmmPixelList toChipCoord(int16 *x = 0,
							 int16 *y = 0);

	XmmPixelList toChipCoord(CalInt16Vector &x,
							 CalInt16Vector &y);

	virtual XmmCartesianPointList &toCamCoord1();
	virtual XmmCartesianPointList &toCamCoord2();

	virtual XmmCartesianPointList &toSacCoord();

	virtual XmmRowlandPointList toRowCoord(real64 *beta = 0,
										   real64 *chi = 0);
	virtual XmmRowlandPointList toRowCoord(CalReal64Vector &beta,
										   CalReal64Vector &chi);

	virtual XmmTelescopePointList toTelCoord(real64 *theta = 0,
											 real64 *phi = 0);
	virtual XmmTelescopePointList toTelCoord(CalReal64Vector &theta,
											 CalReal64Vector &phi);


	virtual std::vector<Ccd> onCcd();
		//
		//	PURPOSE:
		//		Determine for all pixels in the list on which CCD they
		//		lie; return result as vector with CCD numbers; for positions
		//		outside the detector or between chip gaps the value
		//		Xmm::CCD_NONE is returned.
		//


	virtual CalReal64Vector rowlandDefocusLength(double theta = 0.,
												 double phi = 0.);
		//
		//	PURPOSE:
		//		Compute for each point in the list the distance to the
		//		intersection point with the Rowland circle along the
		//		line towards the G-point (see [2] for details)
		//
		//	ARGUMENTS:
		//		theta/phi:	off-axis/azimuth angle in TELCOORD frame ["/rad]
		//
		//	RETURNS:
		//		Vector with defocus lengths in [mm]
		//


	virtual ~XmmCartesianPointList() {}

private:
	CoordFrame				_frame;
	CoordFrameQuantities	_cfq;
	CartesianVector			_cc1ToCc2;

	CartesianVector			_r0;
	unsigned				_ocbX, _ocbY;
	double					_ocbrX, _ocbrY,
							_halfCentralOcbCeilX,
							_halfCentralOcbCeilY,
							_chipExpansion,
							_camExpansion;

	int16 xIntCoord(double x);
	int16 yIntCoord(double y);

	XmmCartesianPointList &toCamCoord2FromSacCoord();
	XmmPixelList toChipCoordSingleCcd(int16 *x, int16 *y);
}; // XmmCartesianPointList()


//
//	---( XmmTelescopePointList )-----------------------------------------------
//
class XmmTelescopePointList : public AngularPointList<CalReal64Vector>,
							  public virtual CalChild<GeometryDataServer> {
public:
	enum CoordFrame { TELCOORD };
	XmmTelescopePointList(CalReal64Vector &theta,
						  CalReal64Vector &phi,
						  CoordFrame frame = TELCOORD,
						  GeometryDataServer *parent = 0);

	virtual XmmCartesianPointList toCamCoord2(real64 *x = 0,
											  real64 *y = 0,
											  real64 *z = 0);
	virtual XmmCartesianPointList toCamCoord2(CalReal64Vector &x,
											  CalReal64Vector &y,
											  CalReal64Vector &z);

	virtual ~XmmTelescopePointList() {}

private:
	CoordFrame				_frame;
}; // XmmTelescopePointList()


//
//	---( RowlandBeta )-------------------------------------------------------
//
class RowlandBeta {
public:
	RowlandBeta(double beta) : _beta(beta) {}

	double beta() const { return _beta; }

	double x() const;
		//
		//	RETURNS:
		//		SACCOORD x-coordinate corresponding to beta angle
		//

	~RowlandBeta() {}

private:
	double	_beta;

	static	InterpolatedCurve *_betaToX;
	static  Xmm::Instrument	   _ins;
	void	init() const;
}; // RowlandBeta()


//
//	---( XmmRowlandPointList )-----------------------------------------------
//
class XmmRowlandPointList : public AngularPointList<CalReal64Vector>,
							public virtual CalChild<GeometryDataServer> {
public:
	enum CoordFrame { ROWCOORD };
	XmmRowlandPointList(CalReal64Vector &beta,
						CalReal64Vector &chi,
						CoordFrame frame = ROWCOORD,
						GeometryDataServer *parent = 0);

	virtual CalReal64Vector &beta() { return theta(); }
	virtual CalReal64Vector &chi() { return phi(); }

	virtual XmmCartesianPointList toSacCoord(real64 *x = 0,
											 real64 *y = 0,
											 real64 *z = 0);
	virtual XmmCartesianPointList toSacCoord(CalReal64Vector &x,
											 CalReal64Vector &y,
											 CalReal64Vector &z);

	virtual CalReal64Vector defocusLength(double theta = 0.,
										  double phi = 0.);
		//
		//	PURPOSE:
		//		Compute for each point in the list the distance to the
		//		intersection point with the Rowland circle along the
		//		line towards the G-point (see [2] for details)
		//
		//	ARGUMENTS:
		//		theta/phi:	off-axis/azimuth angle in TELCOORD frame ["/rad]
		//
		//	RETURNS:
		//		Vector with defocus lengths in [mm]
		//


	virtual ~XmmRowlandPointList() {}

private:
	CoordFrame				_frame;
}; // XmmTelescopePointList()


class FieldEvaluator {
public:
	FieldEvaluator(const string &region);

	virtual bool insideRegion(double rawX, double rawY) const;

	virtual ~FieldEvaluator();

private:
	Evaluator *_evaluator;
	double	  *_pchipX, *_pchipY;
	double	  _chipDum;
};


//
//	---( GeometryDataServer )-----------------------------------------------
//
class GeometryDataServer : public CalAtom,
			 public virtual CalChild<CalAtomCollection<GeometryDataServer> > {
public:
	GeometryDataServer(const CoordFrameQuantities &cfq,
					   const Proxy<FieldEvaluator> &fieldEvaluator,
					 CalAtomCollection<GeometryDataServer> *parent = 0)
		: CalChild<CalAtomCollection<GeometryDataServer> >(parent),
		  _cfq(cfq), _fieldEvaluator(fieldEvaluator) {}

	static unsigned id() { return Ccf::LinCoord; }

	const CoordFrameQuantities &coordFrameQuantities() const
		{ return _cfq; }

	const Proxy<FieldEvaluator> &fieldEvaluator() const
		{ return _fieldEvaluator; }

	const Proxy<std::vector<FBox> > &ccdBoundingBoxes() const
		{ return _ccdBoundingBoxes; }

	virtual void dump(std::ostream &) {}
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//

	XmmPixelList pixCoord0List(CalInt16Vector &x,
							   CalInt16Vector &y)
		{ return XmmPixelList(x, y, XmmPixelList::PIXCOORD0, this); }

	XmmPixelList pixCoord1List(CalInt16Vector &x,
							   CalInt16Vector &y)
		{ return XmmPixelList(x, y, XmmPixelList::PIXCOORD1, this); }

	XmmPixelList chipCoordList(CalInt16Vector &x,
							   CalInt16Vector &y)
		{ return XmmPixelList(x, y, XmmPixelList::CHIPCOORD, this); }

	XmmRealPixelList chipCoordList(CalReal64Vector &x,
								   CalReal64Vector &y)
		{ return XmmRealPixelList(x, y, XmmPixelList::CHIPCOORD, this); }

	XmmCartesianPointList camCoord1List(CalReal64Vector &x,
										CalReal64Vector &y,
										CalReal64Vector &z)
		{ return XmmCartesianPointList(x, y, z,
									XmmCartesianPointList::CAMCOORD1, this); }

	XmmCartesianPointList camCoord2List(CalReal64Vector &x,
										CalReal64Vector &y,
										CalReal64Vector &z)
		{ return XmmCartesianPointList(x, y, z,
									XmmCartesianPointList::CAMCOORD2, this); }

	XmmCartesianPointList sacCoordList(CalReal64Vector &x,
									   CalReal64Vector &y,
									   CalReal64Vector &z)
		{ return XmmCartesianPointList(x, y, z,
									XmmCartesianPointList::SACCOORD, this); }

	XmmTelescopePointList telCoordList(CalReal64Vector &theta,
									   CalReal64Vector &phi)
		{ return XmmTelescopePointList(theta, phi,
									XmmTelescopePointList::TELCOORD, this); }

	XmmRowlandPointList rowCoordList(CalReal64Vector &beta,
									 CalReal64Vector &chi)
		{ return XmmRowlandPointList(beta, chi,
									XmmRowlandPointList::ROWCOORD, this); }
		//
		//	PURPOSE:
		//		Return instance of XmmPixelList/XmmCartesianPointList/
		//		XmmTelescopePointList/XmmRowlandPointList constructed
		//		from respective input vectors in the designated frame
		//
		//	ARGUMENTS:
		//		Input vectors in the appropriate coordinate frame
		//


	Xmm::Ccd onCcd(double x, double y);
	inline bool isOnAnyCcd(double x, double y)
		{ return onCcd(x, y) != Xmm::CCD_NONE; }
		//
		//	PURPOSE:
		//		isOnAnyCcd(): convenience function to test whether a given
		//					  CAMCOORD2 position lies on any CCD
		//		onCcd()		: returns the number of the CCD that comprises
		//					  a given CAMCOORD2 location (x, y) or
		//					  Xmm::CCD_NONE if the location does not lie on
		//					  any CCD (e.g. in a chip gap)
		//


	Spectrum lsfDefocusDistribution(double nBetaHWHM,
									double betaChannelWidth,
									double theta,
									double phi,
									int order,
									double E);
		//
		//	PURPOSE:
		//		Compute broadening component of the RGS Line-Spread-Function
		//		that is due to the defocusing of the RFC. This is a
		//		consequence of the fact that the flat RFC CCD surfaces
		//		do not exactly match the Rowland circle geometry. This
		//		is thus a purely geometrical effect. See [2] for details.
		//
		//	ARGUMENTS:
		//		nBetaHWHM	: number of Half-Width-Half-Maximum units to
		//				      evaluate the distribution to; this determines
		//					  the number of channels of the output spectrum
        //      betaChannelWidth:
        //                    width of channels of beta grid [rad]
		//		theta/phi	: off-axis/azimuth angle [arcsec/rad] of source
		//					  in FOV
		//		order		: spectral order index [unitless] in range [-5, 0]
		//		E			: energy of the incident photon [eV]
		//
		//	RETURNS:
		//		Normalized probability distribution in beta-space. The first
		//		moment of the distribution coincides with the central
		//		spectrum channel. Please note: The distribution is not
		//		in 'wrap-around' (such that its first moment would lie in
		//		channel 0). This is needed to convolve the distribution with
		//		other components to form the LSF. The Spectrum class has
		//		two methods
		//			* Spectrum::firstMoment()
		//			* Spectrum::shift()
		//		to perform this operation efficiently outside
		//		GratingDataServer::lsfDefocusDistribution()
		//


	~GeometryDataServer() {}
		//
		//	PURPOSE:
		//		Destroy GeometryDataServer object instantiated with (one)
		//		of above constructors
		//

private:
	CoordFrameQuantities	_cfq;	//	quantities need in various frame
									//	conversions
	Proxy<FieldEvaluator>	_fieldEvaluator;
									//	object capable of determining whether
									//	or not a given position is within the
									//	FOV
	Proxy<std::vector<FBox> >	_ccdBoundingBoxes;
									//	rectangle vector defining the CCDs
	friend class CalAtomCollection<GeometryDataServer>;
	void registerCcdBoxes(const Proxy<std::vector<FBox> > &ccdBoundingBoxes)
	{ _ccdBoundingBoxes = ccdBoundingBoxes; }
}; // GeometryDataServer

#endif
