//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	OmFluxConverter.cc --- implementation of OmFluxConverter class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 2000-01-27: 1.0.0: - initial version
//		UL, 2000-03-03: 1.1.0: - OmFluxConverter::fluxCorrect(): brought
//								 implementation in line with latest
//								 description in CAL HB
//		UL, 2002-01-28: 1.2.0: - fluxCorrect(): estimate errors of converted
//								 count rate (does not take into account
//								 coincidence loss correction)
//		UL, 2002-04-24: 1.3.0: - fixed bug in calculation of converted count
//								 rate
//
//	AUTHOR:
//		U. Lammers (UL), Thu Jan 27 11:39:07 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, XMM instrument team
//
//	$Header: /home/sasbuild/repositories/sasdev/sas/packages/cal/src/OmFluxConverter.cc,v 1.15 2003/05/23 15:01:12 ulammers Exp $
//

#include <cfloat>
#include <cmath>
#include <climits>

#include "OmFluxConverter.h"
#include "CalOalUtility.h"

using namespace std;

//
//	---( OmFluxConverter> ) ---------------------------------------
//
OmFluxConverter::OmFluxConverter(const CcfConstituentTable &photToNatTab,
					CalAtomCollection<OmFluxConverter> *parent)
	: CalChild<CalAtomCollection<OmFluxConverter> >(parent)
{
	init(photToNatTab);
} // OmFluxConverter::OmFluxConverter()


OmFluxConverter::OmFluxConverter(const CcfConstituentTable &photToNatTab,
					const CcfConstituentTable &coincidenceTab,
					CalAtomCollection<OmFluxConverter> *parent)
	: CalChild<CalAtomCollection<OmFluxConverter> >(parent)
{
	init(photToNatTab, &coincidenceTab);
} // OmFluxConverter::OmFluxConverter()


void OmFluxConverter::init(const CcfConstituentTable &photToNatTab,
						   const CcfConstituentTable *coincidenceTab)
{
	const Table *table = photToNatTab.block();

	//
	//	read attributes
	//
	_phdFrac = table->attribute("PHDFRAC")->asReal(),
	_thresh	= table->attribute("THRESH")->asReal();

	//
	//	read table
	//
	ColumnData	*cdparam = table->column("PARAM00")->data(),
				*cdparame = table->column("DPARAM00")->data();
	real32 *param = cdparam->real32Data(),
		   *parame = cdparame->real32Data();

	if (cdparam->elements() < 8)
		errstr<<"Need a least 8 elements in column "
			  <<table->column("PARAM00")->qualifiedName()
			  <<Sas::error("TooLittleData");
	if (cdparam->elements() != cdparame->elements())
		errstr<<"Number of elements in columns "
			  <<table->column("PARAM00")->qualifiedName()
			  <<" and "
			  <<table->column("DPARAM00")->qualifiedName()
			  <<" do not match"
			  <<Sas::error("VectorLengthMismatch");

	OmFluxConverter::FrameParameters &fp = _frameParameters;
	fp.nHorizontalPixels.value = param[cdparam->position(0, 0)];
	fp.nHorizontalPixels.error = parame[cdparame->position(0, 0)];
	fp.nVerticalPixels.value = param[cdparam->position(0, 1)];
	fp.nVerticalPixels.error = parame[cdparame->position(0, 1)];
	fp.verticalTransferTime.value = param[cdparam->position(0, 2)];
	fp.verticalTransferTime.error = parame[cdparame->position(0, 2)];
	fp.horizontalTransferTime.value = param[cdparam->position(0, 3)];
	fp.horizontalTransferTime.error = parame[cdparame->position(0, 3)];
	fp.nVerticalPixelsImage.value = param[cdparam->position(0, 4)];
	fp.nVerticalPixelsImage.error = parame[cdparame->position(0, 4)];
	fp.verticalHorizontalSwitchTime.value = param[cdparam->position(0, 5)];
	fp.verticalHorizontalSwitchTime.error = parame[cdparame->position(0, 5)];
	fp.horizontalVerticalSwitchTime.value = param[cdparam->position(0, 6)];
	fp.horizontalVerticalSwitchTime.error = parame[cdparame->position(0, 6)];
	fp.redundantCcdOffset.value = param[cdparam->position(0, 7)];
	fp.redundantCcdOffset.error = parame[cdparame->position(0, 7)];

	//	tidy up
	delete cdparam;
	delete cdparame;

	//	read COINCIDENCE table if present
	if (coincidenceTab)
		CalOalUtility::fillVector(coincidenceTab->block(), "PLINFUNC", 0,
								  _linearityCoeff);
} // OmFluxConverter::init()


void OmFluxConverter::dump(ostream &os)
{
	os<<"OmFluxConverter{\n";
	os<<"}";
} // OmFluxConverter::dump()


OmFluxConverter::FrameParameters OmFluxConverter::frameParameters() const
{
	return _frameParameters;
} // OmFluxConverter::frameParameters()


double OmFluxConverter::phdFrac() const
{
	return _phdFrac;
} // OmFluxConverter::phdFrac()


double OmFluxConverter::thresh() const
{
	return _thresh;
} // OmFluxConverter::thresh()



ValueAsymmetricError OmFluxConverter::fluxCorrect(ValueAsymmetricError rawRate,
									  double frameTime,
									  double deadFrac,
									  double eventLoss,
									  bool extendedSource,
									  bool empiricalLinearityCorrection) const
{
	double rate = rawRate.value;
	ValueAsymmetricError res = rawRate;

	//
	//	coincidence loss correction (only for non-extended sources)
	//
	if (!extendedSource) {
		if (frameTime == 0.)
			errstr<<"Frametime is 0 - cannot correct raw OM count rate"
				  <<Sas::error("IllegalFrameTime");

		if (rate < 1./frameTime) {
			double f = 1./(1.-rate*frameTime);
			rate = -log(1.-rate*frameTime)/frameTime;
			res.error.errorPlus = f*rawRate.error.errorPlus;
			res.error.errorMinus = f*rawRate.error.errorMinus;
		} else {
			res.value = DBL_MAX;
			res.error.errorPlus = res.error.errorMinus = DBL_MAX;
			errstr<<"The count rate is greater than the frame rate - this is "
					"not possible - cannot correct raw OM count rate"
			  <<Sas::warning("IllegalCountRate");
			return res;
		}

		//	apply empirical correction if requested
		unsigned N = _linearityCoeff.size();
		if (empiricalLinearityCorrection && N) {
			unsigned j=N-1;
			double v = _linearityCoeff[j],
				   x = rawRate.value * frameTime;
			while (j>0)
				v = v*x + _linearityCoeff[--j];
			rate *= v;
		}
	}

	//
	//	correct for deadtime
	//
	if (deadFrac == 1.)
		errstr<<"Deadtime is 1 - cannot correct raw OM count rate"
			  <<Sas::error("IllegalDeadTime");
	rate /= 1.-deadFrac;

	//
	//	correction for average detector aging
	//
	if (eventLoss != 0.)
		if (eventLoss == 1.)
			errstr<<"Fraction of events below threshold is 1 - cannot "
					"correct raw OM count rate"
				  <<Sas::error("IllegalEventLossRate");
		else
			rate /= (1.-eventLoss) * _phdFrac;

	res.value = rate;

	return res;
} // OmFluxConverter::fluxCorrect()


OmFluxConverter::~OmFluxConverter()
{
} // OmFluxConverter::~OmFluxConverter()


//
//	---( CalAtomCollection<OmFluxConverter> ) ----------------------
//
CalAtomCollection<OmFluxConverter>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	OmFluxConverter *null = 0;
	_collection = new vector<OmFluxConverter *>(_nEntries, null);
	OmFluxConverter *&ff = (*_collection)[0];

	//
	//	access data tables
	//
	CifEntry *e = ccf->entry(instrument, Ccf::PhotToNat);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, "PHOTTONAT");
	if (!dataSet->hasBlock("COINCIDENCE")) {
		errstr<<"Data set `"<<dataSet->qualifiedName()<<"' does not possess "
				"a table COINCIDENCE - the empirical rate linearity correction"
				" will have no effect"
			  <<Sas::warning("MissingCoincidenceTable");
		ff = new OmFluxConverter(ccftab, this);
	} else
		ff = new OmFluxConverter(ccftab,
								 CcfConstituentTable(dataSet, "COINCIDENCE"),
								 this);

	dataSetServer->close(dataSet);
} // CalAtomCollection<OmFluxConverter>::CalAtomCollection()


OmFluxConverter *CalAtomCollection<OmFluxConverter>::select(
					Ccd,
					Node,
					unsigned) const
{
	return (*_collection)[0];
} // CalAtomCollection<OmFluxConverter>::select
